%% Embedding Nonlinear Mass-Spring-Damper System
%
% 02/2018 Maarten Schoukens

clear all
close all
clc

%% variables
N = 2^14;
M = 2;
P = 2;
Ptr = 3;
Ptot = P+Ptr;

fs = 10; %Hz - sample time of the data record
tTot = [0:1/fs:(N*Ptot-1)/fs];
tP = [0:1/fs:(N-1)/fs];
freq = 0:fs/N:fs-fs/N;

lines = 2:round(N/5);
nLines = length(lines);

uStd = .3;

%% system description - continuous time

% s^2*y(t) + c1*s*y(t) + k1*y(t) = u(t) - (c2*sin(y(t))*(s*y(t)) + c3*(s*y(t))^2 + k2*y(t)^3)
%
% 1-DOF mass-spring-damper system with unity mass and:
% position dependent damping: c2*sin(y(t))*(s*y(t))
% nonlinear (quadratic) damping: c3*(s*y(t))^2
% nonlinear cubic spring: k2*y(t)^3

% system parameters
nx = 4;
nu = 2;
ny = 2;
nz = 2;
nw = 1;

c1 = .1;
c2 = .01;
k1 = (0.5*2*pi)^2;
k2 = (0.6*2*pi)^2;

c3 = .1;
c4 = .2;
k3 = 10;

C = [c1+c2 -c2; -c2 c2];
K = [k1+k2 -k2; -k2 k2];
% M is a identity matrix

Ass = [zeros(nu) eye(nu); -K -C];
Bss = [zeros(nu); eye(nu)];
Css = [eye(ny) zeros(ny)];
Dss = [zeros(ny,nu)];

Bnl = [0 0 -1 0]';

x0 = [0;0];

% Simulink implementation
% see NlMassSpringDamper.slx

%% embedding

Alpv0 = Ass;
Blpv0 = Bss;
Clpv0 = Css;
Dlpv0 = Dss;

Bw = [0 0 -1 0]';
Cz = [1 0 0 0; 0 0 1 0];

% Alpv1 = [0 0 0 0; 0 0 0 0; -p1 0 -p2 0; 0 0 0 0];

hp1 = @(y) k2*y.^2;
hp2 = @(y,ydot) c2*sin(y) + c3*ydot;

%% simulation

% generate input
% u = repmat(uStd*randn(N,1),[Ptot,1]);

U = zeros(N,nu,M);
U(lines,:,:) = exp(1i*2*pi*randn(nLines,nu,M));
u = real(ifft(U));
u = uStd*u./std(u(:,1,1));
u = repmat(u,[Ptot,1]);

% put 2nd input to zero
u(:,2,:) = 0;

y = zeros([N*Ptot ny M]);
x = zeros([N*Ptot nx M]);
xNL = zeros([N*Ptot nx M]);
% Nonlinear Mass Spring Damper Simulation
for ii=1:M
    input = timeseries(u(:,:,ii),tTot);
    input = setinterpmethod(input,'zoh');

    sim('NlMassSpringDamper2DOF.slx');

    y(:,:,ii) = squeeze(simOut.Data);
    x(:,:,ii) = squeeze(simStates.Data);
    xNL(:,:,ii) = squeeze(simNL.Data);
end

% LPV simulation
yLPV = zeros(size(u));
xLPV = zeros([N*Ptot nx M]);
Alpv1 = zeros([nx nx N*Ptot M]);
for ii=1:M    
    input = timeseries(u(:,:,ii),tTot);
    input = setinterpmethod(input,'zoh');

    sim('NlMassSpringDamper2DOF_LPV_v2.slx');

    yLPV(:,:,ii)   = squeeze(simOut.Data);
    xLPV(:,:,ii) = squeeze(simStates.Data);    
    Alpv1(:,:,:,ii) = squeeze(simAlpv1.Data);    
end


% remove transient
u = u(N*Ptr+1:end,:,:);
y = y(N*Ptr+1:end,:,:);
x = x(N*Ptr+1:end,:,:);
xNL = xNL(N*Ptr+1:end,:,:);
yLPV = yLPV(N*Ptr+1:end,:,:);
xLPV = xLPV(N*Ptr+1:end,:,:);
Alpv1 = Alpv1(:,:,N*Ptr+1:end,:);
%% plots

% BLA analysis plots
% G1 = fft(y(end-N+1:end,1,:))./fft(u(end-N+1:end,1,:));
% G2 = fft(y(end-N+1:end,2,:))./fft(u(end-N+1:end,1,:));
% GBla1 = mean(G1(lines,:),2);
% GBla2 = mean(G2(lines,:),2);
% GVar1 = var(G1(lines,:),[],2);
% GVar2 = var(G2(lines,:),[],2);
% 
% figure; hold on;
% plot(freq(lines),db(GBla1),'b')
% plot(freq(lines),db(GVar1,'power'),'b*')
% plot(freq(lines),db(GBla2),'r')
% plot(freq(lines),db(GVar2,'power'),'r*')
% legend('G_{BLA}','var_G')
% xlabel('freq (Hz)')
% ylabel('magnitude (dB)')

% LPV embedding plots - time domain
figure; 
subplot(2,1,1); hold on;
plot(tP,y(end-N+1:end,1,1),'o')
plot(tP,yLPV(end-N+1:end,1,1))
plot(tP,y(end-N+1:end,1,1)-yLPV(end-N+1:end,1,1),'*')
xlabel('time (s)'); ylabel('position (m)'); title('y_1(t)')
axis tight; grid on
subplot(2,1,2); hold on;
plot(tP,y(end-N+1:end,2,1),'o')
plot(tP,yLPV(end-N+1:end,2,1))
plot(tP,y(end-N+1:end,2,1)-yLPV(end-N+1:end,2,1),'*')
xlabel('time (s)'); ylabel('position (m)'); title('y_2(t)')
axis tight; grid on
% LPV embedding plots - frequency domain
figure; hold on;
plot(freq,db(1/N*fft(y(end-N+1:end,1,1))),'o')
plot(freq,db(1/N*fft(yLPV(end-N+1:end,1,1))))
plot(freq,db(1/N*fft(y(end-N+1:end,2,1))),'o')
plot(freq,db(1/N*fft(yLPV(end-N+1:end,2,1))))
xlabel('frequency (Hz)'); ylabel('magnitude (dB)');
v = axis; axis([freq(lines(1)) freq(lines(end)) -120 -20])
grid on;



% NL plots
% scheduling 1
figure; plot(squeeze(xLPV(:,1,1)),squeeze(Alpv1(3,1,:,1)),'.');
xlabel('x_1'); ylabel('p_1'); grid on
title('f_1(x_1)')
% scheduling 2
figure; scatter3(squeeze(xLPV(:,1,1)),squeeze(xLPV(:,3,1)),squeeze(Alpv1(3,3,:,1)),5,squeeze(Alpv1(3,3,:,1)));
colormap(jet); colorbar; grid on
xlabel('x_1'); ylabel('x_3'); zlabel('p_2')
title('f_2(x_1,x_3)')


% multivariate NL
figure; hold on;
scatter3(squeeze(x(:,1,1)),squeeze(x(:,3,1)),squeeze(xNL(:,3,1)),5,squeeze(xNL(:,3,1)));
% scatter3(squeeze(x(:,1,1)),squeeze(x(:,3,1)),squeeze(Alpv1(3,1,:,1)).*squeeze(x(:,1,1))+squeeze(Alpv1(3,3,:,1)).*squeeze(x(:,3,1)),'o');
colormap(jet); colorbar; grid on
xlabel('z_1'); ylabel('z_2'); zlabel('f(z_1,z_2)')

%% to save figures as pdf
h = gcf;set(h,'Units','Inches');
pos = get(h,'Position');
set(h,'PaperPositionMode','Auto','PaperUnits','Inches','PaperSize',[pos(3), pos(4)])
print(h,'SimExample_NL','-dpdf','-r0')

addpath(genpath('..\Functions Extern\Fig3DPDf'))
